#include <vector>
#include <time.h>
#include <iostream>

float float_EuclideanDistance(float const *lhs, float const *rhs){
        float ans = 0.0;
        unsigned lensDim = 100;
        for (unsigned i = 0; i < lensDim; ++i) {
            float v = float(lhs[i]) - float(rhs[i]);
            v *= v;
            ans += v;
        }
        return ans;
}

int int_EuclideanDistance(int const *lhs, int const *rhs){
        int ans = 0;
        unsigned lensDim = 100;
        for (unsigned i = 0; i < lensDim; ++i) {
            int v = int(lhs[i]) - int(rhs[i]);
            v *= v;
            ans += v;
        }
        return ans;
}

int int_EuclideanDistance(unsigned char const *lhs, unsigned char const *rhs){
        int ans = 0;
        unsigned lensDim = 100;
        for (unsigned i = 0; i < lensDim; ++i) {
            int v = int(lhs[i]) - int(rhs[i]);
            v *= v;
            ans += v;
        }
        return ans;
}

unsigned int int_EuclideanDistance(unsigned short const *lhs, unsigned short const *rhs){
        unsigned int ans = 0;
        unsigned lensDim = 100;
        for (unsigned i = 0; i < lensDim; ++i) {
            unsigned int v = lhs[i] - rhs[i];
            v *= v;
            ans += (unsigned int)v;
        }
        return ans;
}

#include <xmmintrin.h>
#define SSE_L2SQR(addr1, addr2, dest, tmp1, tmp2) \
    tmp1 = _mm_load_ps(addr1);\
    tmp2 = _mm_load_ps(addr2);\
    tmp1 = _mm_sub_ps(tmp1, tmp2); \
    tmp1 = _mm_mul_ps(tmp1, tmp1); \
    dest = _mm_add_ps(dest, tmp1); 
float float_l2sqr_sse2 (float const *t1, float const *t2, unsigned dim) {
    //time_t start,end;
    //start = clock();
    __m128 sum;
    __m128 l0, l1, l2, l3;
    __m128 r0, r1, r2, r3;
    unsigned D = (dim + 3) & ~3U;
    unsigned DR = D % 16;
    unsigned DD = D - DR;
    const float *l = t1;
    const float *r = t2;
    const float *e_l = l + DD;
    const float *e_r = r + DD;
    float unpack[4] __attribute__ ((aligned (16))) = {0, 0, 0, 0};
    float ret = 0.0;
    sum = _mm_load_ps(unpack);
    switch (DR) {
        case 12:
            SSE_L2SQR(e_l+8, e_r+8, sum, l2, r2);
        case 8:
            SSE_L2SQR(e_l+4, e_r+4, sum, l1, r1);
        case 4:
            SSE_L2SQR(e_l, e_r, sum, l0, r0);
    }
    for (unsigned i = 0; i < DD; i += 16, l += 16, r += 16) {
        SSE_L2SQR(l, r, sum, l0, r0);
        SSE_L2SQR(l + 4, r + 4, sum, l1, r1);
        SSE_L2SQR(l + 8, r + 8, sum, l2, r2);
        SSE_L2SQR(l + 12, r + 12, sum, l3, r3);
    }
    _mm_storeu_ps(unpack, sum);
    ret = unpack[0] + unpack[1] + unpack[2] + unpack[3];
    //end=clock();
    //std::cout<<"SSE2"<<end-start<<std::endl;
    return ret;
}

#include <immintrin.h>
#define AVX_L2SQR(addr1, addr2, dest, tmp1, tmp2) \
    tmp1 = _mm256_loadu_ps(addr1);\
    tmp2 = _mm256_loadu_ps(addr2);\
    tmp1 = _mm256_sub_ps(tmp1, tmp2); \
    tmp1 = _mm256_mul_ps(tmp1, tmp1); \
    dest = _mm256_add_ps(dest, tmp1); 
float float_l2sqr_avx (float const *t1, float const *t2, unsigned dim) {
    __m256 sum;
    __m256 l0, l1, l2, l3;
    __m256 r0, r1, r2, r3;
    unsigned D = (dim + 7) & ~7U; // # dim aligned up to 256 bits, or 8 floats
    unsigned DR = D % 32;
    unsigned DD = D - DR;
    const float *l = t1;
    const float *r = t2;
    const float *e_l = l + DD;
    const float *e_r = r + DD;
    float unpack[8] __attribute__ ((aligned (32))) = {0, 0, 0, 0, 0, 0, 0, 0};
    float ret = 0.0;
    sum = _mm256_load_ps(unpack);
    switch (DR) {
        case 24:
            AVX_L2SQR(e_l+16, e_r+16, sum, l2, r2);
        case 16:
            AVX_L2SQR(e_l+8, e_r+8, sum, l1, r1);
        case 8:
            AVX_L2SQR(e_l, e_r, sum, l0, r0);
    }
    for (unsigned i = 0; i < DD; i += 32, l += 32, r += 32) {
        AVX_L2SQR(l, r, sum, l0, r0);
        AVX_L2SQR(l + 8, r + 8, sum, l1, r1);
        AVX_L2SQR(l + 16, r + 16, sum, l2, r2);
        AVX_L2SQR(l + 24, r + 24, sum, l3, r3);
    }
    _mm256_storeu_ps(unpack, sum);
    ret = unpack[0] + unpack[1] + unpack[2] + unpack[3]
        + unpack[4] + unpack[5] + unpack[6] + unpack[7];
    return ret;//sqrt(ret);
}

#include <immintrin.h>
#define AVX512_L2SQR(addr1, addr2, dest, tmp1, tmp2) \
    tmp1 = _mm512_load_ps(addr1);\
    tmp2 = _mm512_load_ps(addr2);\
    tmp1 = _mm512_sub_ps(tmp1, tmp2); \
    tmp1 = _mm512_mul_ps(tmp1, tmp1); \
    dest = _mm512_add_ps(dest, tmp1); 
float float_l2sqr_avx512 (float const *t1, float const *t2, unsigned dim) {
    __m512 sum;
    __m512 l0, l1, l2, l3;
    __m512 r0, r1, r2, r3;
    unsigned D = (dim + 15) & ~15U; // # dim aligned up to 256 bits, or 8 floats
    unsigned DR = D % 64;
    unsigned DD = D - DR;
    const float *l = t1;
    const float *r = t2;
    const float *e_l = l + DD;
    const float *e_r = r + DD;
    float unpack[16] __attribute__ ((aligned (64))) = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    float ret = 0.0;
    sum = _mm512_load_ps(unpack);
    switch (DR) {
        case 48:
            AVX512_L2SQR(e_l+32, e_r+32, sum, l2, r2);
        case 32:
            AVX512_L2SQR(e_l+16, e_r+16, sum, l1, r1);
        case 16:
            AVX512_L2SQR(e_l, e_r, sum, l0, r0);
    }
    for (unsigned i = 0; i < DD; i += 64, l += 64, r += 64) {
        AVX512_L2SQR(l, r, sum, l0, r0);
        AVX512_L2SQR(l + 16, r + 16, sum, l1, r1);
        AVX512_L2SQR(l + 32, r + 32, sum, l2, r2);
        AVX512_L2SQR(l + 48, r + 48, sum, l3, r3);
    }
    _mm512_storeu_ps(unpack, sum);
    ret = unpack[0] + unpack[1] + unpack[2] + unpack[3]
        + unpack[4] + unpack[5] + unpack[6] + unpack[7]
        + unpack[8] + unpack[9] + unpack[10] + unpack[11]
        + unpack[12] + unpack[13] + unpack[14] + unpack[15];
    return ret;//sqrt(ret);
}

#define INT_AVX512_L2SQR(addr1, addr2, dest, tmp1, tmp2) \
    tmp1 = _mm512_load_epi32(addr1);\
    tmp2 = _mm512_load_epi32(addr2);\
    tmp1 = _mm512_sub_epi32(tmp1, tmp2); \
    tmp1 = _mm512_mul_epi32(tmp1, tmp1); \
    dest = _mm512_add_epi32(dest, tmp1); 
int int_l2sqr_avx512 (int const *t1, int const *t2, unsigned dim) {
    __m512i sum;
    __m512i l0, l1, l2, l3;
    __m512i r0, r1, r2, r3;
    unsigned D = (dim + 15) & ~15U; // # dim aligned up to 256 bits, or 8 floats
    unsigned DR = D % 64;
    unsigned DD = D - DR;
    const int *l = t1;
    const int *r = t2;
    const int *e_l = l + DD;
    const int *e_r = r + DD;
    int unpack[16] __attribute__ ((aligned (64))) = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    int ret = 0.0;
    sum = _mm512_load_epi32(unpack);
    switch (DR) {
        case 48:
            INT_AVX512_L2SQR(e_l+32, e_r+32, sum, l2, r2);
        case 32:
            INT_AVX512_L2SQR(e_l+16, e_r+16, sum, l1, r1);
        case 16:
            INT_AVX512_L2SQR(e_l, e_r, sum, l0, r0);
    }
    for (unsigned i = 0; i < DD; i += 64, l += 64, r += 64) {
        INT_AVX512_L2SQR(l, r, sum, l0, r0);
        INT_AVX512_L2SQR(l + 16, r + 16, sum, l1, r1);
        INT_AVX512_L2SQR(l + 32, r + 32, sum, l2, r2);
        INT_AVX512_L2SQR(l + 48, r + 48, sum, l3, r3);
    }
    _mm512_storeu_epi32(unpack, sum);
    ret = unpack[0] + unpack[1] + unpack[2] + unpack[3]
        + unpack[4] + unpack[5] + unpack[6] + unpack[7]
        + unpack[8] + unpack[9] + unpack[10] + unpack[11]
        + unpack[12] + unpack[13] + unpack[14] + unpack[15];
    return ret;//sqrt(ret);
}